import json
import requests
from requests_oauthlib import OAuth1
from .globals import kogi_get, globals_update
from .s3logging import kogi_print

# Translate

TEXTRA_NAME = 'kkuramitsu'
TEXTRA_KEY = '6c0bbdfd6c5c53cb0b0699729ed56a5c062ebba7c'
TEXTRA_URL = 'https://mt-auto-minhon-mlt.ucri.jgn-x.jp/api/mt/generalNT'
TEXTRA_CACHE = {}
TEXTRA_OAUTH = None


def load_mt(secret):
    global TEXTRA_OAUTH
    TEXTRA_OAUTH = OAuth1(TEXTRA_KEY, secret)
    return TEXTRA_OAUTH


def check_oauth():
    global TEXTRA_OAUTH
    if TEXTRA_OAUTH is None:
        secret = kogi_get('textra', None)
        if secret:
            return load_mt(secret)
        return None
    return TEXTRA_OAUTH


def _isEnglish(text):
    for c in text:
        if c >= 'あ':
            return False
    return True


def translate(text, lang=None):
    global TEXTRA_OAUTH
    check_oauth()
    if TEXTRA_OAUTH is None:
        return None

    if lang is not None:
        URL = f'{TEXTRA_URL}_{lang}/'
    elif _isEnglish(text):
        URL = f'{TEXTRA_URL}_en_ja/'
    else:
        URL = f'{TEXTRA_URL}_ja_en/'

    isMulti = False
    if isinstance(text, list):
        text = '\n'.join(text)
        isMulti = True

    if text in TEXTRA_CACHE:
        return TEXTRA_CACHE[text]

    params = {
        'key': TEXTRA_KEY,
        'name': TEXTRA_NAME,
        'type': 'json',
        'text': text,
    }

    try:
        res = requests.post(
            URL, data=params, auth=TEXTRA_OAUTH, timeout=(3.5, 7.0))
        res.encoding = 'utf-8'
        data = json.loads(res.text)
        result = data['resultset']['result']['text']
        if isMulti:
            return result.split('\n')
        TEXTRA_CACHE[text] = result
        return result
    except Exception as e:
        kogi_print('翻訳できません', e)
        TEXTRA_OAUTH = None
        globals_update({'textra': None})
        return None


def translate_en(text):
    return translate(text, lang='en_ja')


def translate_ja(text):
    return translate(text, lang='ja_en')



import traceback
from kogi.service import record_log, kogi_set, debug_print
from ._google import google_colab
from .message import kogi_print, Doc
from IPython.display import JSON

LOGIN_HTML = """\
<style>
/* Bordered form */
form {
  border: 3px solid #f1f1f1;
}

/* Full-width inputs */
input[type=text] {
  width: 100%;
  padding: 6px 10px;
  margin: 8px 0;
  display: inline-block;
  border: 1px solid #ccc;
  box-sizing: border-box;
}

/* Set a style for all buttons */
button.login {
  color: white;
  background-color: #f44336;
  margin: 8px 0;
  border: none;
  cursor: pointer;
  width: auto;
  padding: 8px 16px;
  background-color: #f44336;
}

button:disabled {
    background-color: #aaaaaa;
    filter:brightness(0.5);
    cursor:not-allowed;
}

/* Add a hover effect for buttons */
button:hover {
  opacity: 0.8;
}

/* Add padding to containers */
.container {
  padding: 0px;
}

/* The "Forgot password" text */
span.psw {
  float: right;
  padding-top: 8px;
}

</style>
<form id="base">
  <b>こんにちは！ コギーくんは、皆さんの学習状況にあわせてお手伝いします。</b>
  <div class="container">
    <label for="uname">ニックネーム</label>
    <input type="text" placeholder="コギーに呼ばれたい名前を入れてね" id="uname" name="uname" required>
    <label for="psw">タイピング力も見せてね</label>
    <div><code id="code">print("A", "B", "C")</code><div>
    <input type="text" placeholder="上のコードを入力してください。" id="ucode" name="ucode" required>
    <div style="font-size:8pt">コードの意味が同じなら、空白は省略して構いません。</div>
    </div>
  <div class="container" style="background-color:#f1f1f1">
    <button type="button" id="ulogin" class="login">利用規約に同意する</button>
    <span class="psw"> <a href="https://kuramitsulab.github.io/kogi_tos.html" target="_blank">利用規約とは</a></span>
  </div>
</form>
<script>
    const samples = [
        'print("Hello,\\\\nWorld")',
        'print((math.pi * i) / 32)',
        'print("X", 1, "Y", 2, "Z")',
        'print(a[x][y], b[x][y])',
        'print(x if x == y else y)',
        'print(a/gcd(a,b), b/gcd(a,b))',
        'print(file=w, end="")',
        'print(1+2, 2*3, 3//4)',
        'print([1,2,3], (1,2,3))',
        'print({"A": 1, "B": 2})',
    ];
    const index = Math.floor(Math.random() * samples.length);
    document.getElementById('code').innerText=samples[index];
    var buffers = [];
    var before = new Date().getTime();
    var finished = false;
    document.getElementById('ucode').addEventListener('keydown', (e) => {
      var now = new Date().getTime();
      if(e.key === ' ') {
        buffers.push(`${now - before} SPACE`);
      }
      else {
        buffers.push(`${now - before} ${e.key}`);
      }
      if(e.key === ')') {
        finished = true;
      }
      before = now;
      if (finished && buffers.length > samples[index].length) {
        document.getElementById('ulogin').disabled=false;
      }
    });
    document.getElementById('ulogin').disabled=true;
    document.getElementById('ulogin').onclick = () => {
        const uname = document.getElementById('uname').value;
        const ucode = document.getElementById('ucode').value;
        const keys = buffers.join(' ');
        //document.getElementById('code').innerText=keys;
        //google.colab.kernel.invokeFunction('notebook.login', [uname, samples[index], ucode, keys], {});
        (async function() {
            const result = await google.colab.kernel.invokeFunction('notebook.login', [uname, samples[index], ucode, keys], {});
            const data = result.data['application/json'];
            document.getElementById('base').innerText=data.text;
        })();
        document.getElementById('base').innerText='';
    };
</script>
"""


def check_level(ukeys):
    keys = ukeys.split()
    times = [int(t) for t in keys[0::2]]
    keys = keys[1::2]
    average_time = (sum(times)-max(times)) / (len(times)-1)
    if average_time < 300:
        return average_time, 5
    if average_time < 400:
        return average_time, 4
    if average_time < 450:
        return average_time, 3
    if average_time < 500:
        return average_time, 2
    return average_time, 1


ULEVEL = [
    '今日も一緒にがんばりましょう！',
    '今日はとってもプログラミング日和よね！',
    '最近、どんどん上達している感じだね！',
    'なんだか、プログラミングは十分、得意そうだね！',
    'お、上級者来たね！',
]

ZHTBL = str.maketrans('０１２３４５６７８９', '0123456789')
STUDENT_CODE = 'm8YpbzR6ovEjyzJ8oXnpT3BlbkFJYwKQCc1DmrTOj4Adj'
K = 'k'

def ulogin(uname, code, ucode, ukeys):
    try:
        kogi_set(approved=True)
        uname = '名無し子' if uname.strip() == '' else uname
        average_time, ulevel = check_level(ukeys)
        kogi_set(uname=uname, ulevel=ulevel, approved=True)
        record_log(type='key', uname=uname, code=code,
                   ucode=ucode, average_time=average_time,
                   ulevel=ulevel, ukeys=ukeys)
        msg = f'{uname}さん！ {ULEVEL[ulevel-1]}'
        kogi_set(about_me=uname, openai_key=f's{K}-{STUDENT_CODE}Ag8')
        return JSON({'text': msg})
    except:
        traceback.print_exc()
        return JSON({'text': 'よろしく！'})


def login(login_func=ulogin):
    if google_colab:
        google_colab.register_callback('notebook.login', login_func)
    doc = Doc.HTML(LOGIN_HTML)
    doc.set_mention('@ta')
    kogi_print(doc, height=280)



import os
import requests
import socket
import time
from .s3logging import debug_print


def check_port(port):
    host = "127.0.0.1"
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((host, port))
    sock.close()
    return result == 0


def getpid():
    try:
        with open('.PID1789') as f:
            return int(f.read())
    except:
        return None


_MODEL_ID = None
_START_TIME = time.time()


def start_server(restart=False):
    if check_port(5000):
        debug_print('Model server is still running')
        pid = getpid()
        if restart and pid:
            debug_print(f'shutdown server {pid}')
            os.system(f'kill -15 {pid}')
        if not restart:
            return
    script = os.path.abspath(__file__).replace('api', 'serv')
    res = os.system(f'python3 {script} {_MODEL_ID} > /dev/null 2>&1 &')
    debug_print(f'Model sever is starting. {_MODEL_ID} res={res}')
    _START_TIME = time.time()


def load_model(model_id):
    global _MODEL_ID
    if _MODEL_ID == model_id:
        return
    restart = (_MODEL_ID is not None)
    _MODEL_ID = model_id
    start_server(restart=restart)


def check_awake():
    try:
        payload = {"inputs": "おはよう", "max_length": 128, "beam": 1}
        requests.post("http://127.0.0.1:5000/predict",
                      json=payload, timeout=(3.5, 7.0))
        return True
    except requests.Timeout as e:
        debug_print(e)
        if not check_port(5000) and time.time() - _START_TIME > 60*10:
            debug_print('Server is down. Trying to restart.')
            start_server(restart=True)
        return False
    except requests.ConnectionError as e:
        debug_print(e)
        if not check_port(5000) and time.time() - _START_TIME > 60*10:
            debug_print('Server is down. Trying to restart.')
            start_server(restart=True)
        return False
    except Exception as e:
        debug_print(e)
        return False


def tabnl(s):
    return s.replace('<tab>', '    ').replace('<nl>', '\n')


def model_generate(text, max_length=128, beam=1):
    try:
        payload = {"inputs": text, "max_length": max_length, "beam": beam}
        response = requests.post("http://127.0.0.1:5000/predict",
                                 json=payload, timeout=(3.5, 7.0))
        output = response.json()
        # print(text, type(output), output)
        if isinstance(output, (list, tuple)):
            output = output[0]
        if 'outputs' in output and beam > 1:
            return [tabnl(s) for s in output['outputs']]
        return tabnl(output.get('generated_text', ''))
    except Exception as e:
        debug_print(e)
        if not check_port(5000) and time.time() - _START_TIME > 60*10:
            debug_print('Server is down. Trying to restart.')
            start_server(restart=True)
        return f'<status>{e}'



from kogi.service import debug_print, check_awake
from kogi.ui.render import Doc
from kogi.ui.wait_and_ready import wait_for_ready_doc


def status_message(status):
    doc = wait_for_ready_doc(check_ready_fn=check_awake)
    doc.set_mention('@ta')
    return doc



import re
from .common import debug_print

_CODE = re.compile(r'(`[^`]+`)')
# _SVAR = re.compile(r'\<[A-Z]\>')


def _extract_patterns(text, pat):
    return re.findall(pat, text)


def encode(text):
    codec = {}
    for code in _extract_patterns(text, _CODE):
        # if re.search(_SVAR, code):
        #     continue
        key = f'@{id(code)}@'
        text = text.replace(code, key)
        codec[key] = code[1:-1]
    return text, codec


def decode(text, codec):
    for key, code in codec.items():
        text = text.replace(key, code)
    return text


def model_parse(text, kw, commands=None):
    kw = dict(kw)
    args = []
    text, codec = encode(text)
    ss = text.replace('　', ' ').split()
    for s in ss:
        if s.startswith('@'):
            if isinstance(commands, list):
                commands.append(s)
            else:
                args = []
        elif '=' in s:
            k, _, v = s.partition('=')
            kw[k] = decode(v, codec)
        else:
            args.append(decode(s, codec))
    return args, kw


_TASK = {

}


def task(names: str):
    def wrapper(func):
        global _TASK
        for name in names.split():
            if name in _TASK:
                debug_print(f'duplicated task {name}')
            _TASK[name] = func
        return func
    return wrapper


def run_prompt(bot, prompt, kwargs):
    global _TASK
    if prompt in _TASK:
        return _TASK[prompt](bot, kwargs)
    else:
        debug_print('undefined task', prompt)
    return None


# def run_task(bot, text, kw):
#     global _TASK
#     cmds = []
#     args, kw = model_parse(text, kw, cmds)
#     ms = []
#     for cmd in cmds:
#         if cmd in _TASK:
#             ms.append(_TASK[cmd](bot, args, kw))
#         else:
#             debug_print('undefined task', cmd)
#     if len(ms) == 0:
#         debug_print('undefined tasks', cmds)
#         return 'あわわわ'
#     return ms



import re
import os
import numbers
import os
import pandas as pd
from IPython import get_ipython
from .service import model_generate, debug_print

# PEGパーサ
try:
    import pegtree as pg
except ModuleNotFoundError:
    os.system('pip install pegtree')
    import pegtree as pg

_PEG = '''

Start = { 
    (Chunk /  { (!Chunk .)* #Chunk } )*
}

Chunk = 
    / SingleQuote
    / DoubleQuote
    / BackQuoteCode
    / Number
    / Variable

SingleQuote = { '\\'' (!'\\'' .)* '\\'' #String } // ' ... '
DoubleQuote = { '"' (!'"' .)* '"' #String }  // " "
BackQuoteCode = '`' {  (!'`' .)*  #Code } '`' // `1+2`

Number = { '-'? [0-9]+ ('.' [0-9]+)? #Number } // 1.2
Variable = { [A-Za-z_0-9]+ #Name }  // a

'''

_parser = pg.generate(pg.grammar(_PEG))

# 型取得
KOGI_TYPEMAP = {
    'bool': '_結果_',
    'int': '_整数_',
    'float': '_数値_',
    'tuple': '_タプル_',
    'list': '_リスト_',
    'str': '_文字列_',
    'function': '_関数_',
    'ndarray': '_配列_',
    'DataFrame': '_データフレーム_',
}


def get_kogitype(value):
    py_type = type(value).__name__
    kg_type = KOGI_TYPEMAP.get(py_type, None)
    if kg_type is not None:
        return kg_type
    if isinstance(value, numbers.Number):
        return '_数値_'
    if hasattr(value, '__iter__'):
        return '_イテラブル_'
    return f'_結果_'


# みんなが使いそうな変数
FREQUENT_NAMES = {'n': '_整数_', 'N': '_整数_',
                  'kogi': 'コギー', 'Kogi': 'コギー', 'KOGI': 'コギー'}


def get_variable_type(name):
    shell = get_ipython()
    if name in shell.user_ns:
        value = shell.user_ns[name]
        return get_kogitype(value)
    else:  # 変数が未定義な場合
        return FREQUENT_NAMES.get(name, '_結果_')


def eval_code_type(code):
    shell = get_ipython()
    try:
        value = eval(code, shell.user_ns)
        return get_kogitype(value)
    except:
        return '_結果_'


def scan_dataframes():
    column_maps = {}
    dataframe_names = []
    shell = get_ipython()
    user_ns = shell.user_ns
    for name in user_ns:
        # 仮
        if name[0] == "_":
            pass
        else:
            value = user_ns[name]
            if isinstance(value, pd.DataFrame):
                dataframe_names.append(name)
                for column in list(value.columns):
                    column_maps[column] = name
                    # column_maps.setdefault(column, name)
    return dataframe_names, column_maps


def detect_string_type(s):
    content = s[1:-1]  # クオートをとる
    if content.endswith('.csv'):
        return '_CSVファイル_'
    if len(content) == 1:
        return '_文字_'
    return '_文字列_'


def append_map(maps, key, value):
    if key not in maps:
        maps[key] = []
    maps[key].append(value)


def parse(text):
    dataframe_names, column_maps = scan_dataframes()
    # debug_print(dataframe_names, column_maps)
    after_maps = {}
    tree = _parser(text)
    ss = []
    for t in tree:
        token = str(t)
        if t == 'Number':
            kgtype = '_数値_' if '.' in token else '_整数_'
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'String':
            kgtype = detect_string_type(token)
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'Name':
            kgtype = get_variable_type(token)
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'Code':
            kgtype = eval_code_type(token)
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'Chunk':
            for column, dfname in column_maps.items():
                if column in token:
                    append_map(after_maps, '_列名_', f'"{column}"')
                    append_map(after_maps, '_データフレーム_', dfname)
                    append_map(after_maps, '_データ列_', f'{dfname}["{column}"]')
                    token = token.replace(column, '_列名_')
            ss.append(token)
    return ''.join(ss).replace('_', ''), after_maps


# モデル出力→ユーザへの出力


VARPAT = re.compile(r'(_[^_]+_)')
# 漢字ひらがなカタカナ
KPAT = re.compile(
    r'(_*[\u3041-\u309F\u30A1-\u30FF\u2E80-\u2FDF\u3005-\u3007\u3400-\u4DBF\u4E00-\u9FFF\uF900-\uFAFF]+_*)')


def get_kvars(text):
    return re.findall(KPAT, text)


def make_output(text, dic, output_fmt='{}'):
    words = get_kvars(text)
    for word in words:
        if word in dic:
            new_word = dic[word].pop(0)
            text = text.replace(word, output_fmt.format(new_word), 1)
            dic[word].append(new_word)
    return text


def model_transform(text, beam=1, transform_before=parse, transform_after=make_output, output_fmt='{}'):
    # debug_print(text)
    user_input, after_maps = transform_before(text)
    # debug_print(user_input, after_maps)
    if beam == 1:
        generated_text = model_generate(user_input, beam=1)
        if generated_text is not None:
            generated_text = transform_after(
                generated_text, after_maps, output_fmt)
        return generated_text
    outputs = model_generate(user_input, beam=beam)
    outputs = [transform_after(t, dict(after_maps), output_fmt)
               for t in outputs]
    return outputs


def rmt_model_transform(text, cache):
    output_fmt = '<b><font color="red">{}</font></b>'
    ss = []
    for line in text.splitlines():
        if line in cache:
            generated = cache[line]
        else:
            generated = model_transform(text, output_fmt=output_fmt)
            if generated.startswith('<') and '>' in generated:
                _, _, generated = generated.partition('>')
                generated = generated.strip()
            cache[line] = generated
        ss.append(generated)
    return '\n'.join(ss)



from kogi.service import *
from ..OLDconversation import ConversationAI, set_chatbot
from .common import status_message
from .all import run_prompt


class MultitaskAI(ConversationAI):

    def response(self, input_text):
        tag, generated_text = self.generate_transform(input_text)
        # debug_print(input_text, tag, generated_text)
        if tag.startswith('<status>'):
            return status_message(generated_text)
        self.record('@model', input_text, f'{tag}{generated_text}')
        kwargs = dict(user_input=input_text,
                      generated_text=generated_text, **self.slots)
        if tag.startswith('<コード'):
            return run_prompt(self, '@translated_code', kwargs)
        if tag.startswith('<コマンド'):
            debug_print('TODO', tag, generated_text)
            return run_prompt(self, generated_text, kwargs)
        return generated_text


set_chatbot(MultitaskAI())



from .ui._google import google_colab
from .ui.render import Doc
from .liberr import kogi_exc
from .ui.message import display_dialog, append_message
import re
import sys
import traceback
import time
from IPython import get_ipython

from .service import (
    translate, model_prompt, kogi_get,
    record_log, debug_print
)

def remove_comment(code):
    ss = []
    for line in code.splitlines():
        if '"' in line or "'" in line:
            ss.append(line)
        else:
            line, _, _ = line.partition('#')
            ss.append(line)
    return '\n'.join(ss)


class ChatAI(object):
    slots: dict
    chats: dict

    def __init__(self, slots=None):
        self.slots = slots or {}
        self.chats = {}
        self.face_icon = '@kogi_plus'
        self.prev_time = time.time()

    def get(self, key, value):
        return self.slots.get(key, value)

    def update(self, context: dict):
        if context:
            self.slots = dict(context)
        else:
            self.slots = {}
        self.slots['rec_id'] = None
        self.slots['role'] = kogi_get('role', 'Pythonに詳しい友人')
        if kogi_get('about_me', None):
            self.slots['about_me'] = kogi_get('about_me')
        self.slots['tone'] = kogi_get('tone', 'お友達口調で優しく教えてください。')
        debug_print('updating slots', self.slots)

    def difftime(self):
        t = time.time()
        diff = int(t - self.prev_time)
        self.prev_time = t
        return diff

    def likeit(self, rec_id, score):
        if rec_id in self.chats:
            context, prompt, response, data = self.chats[rec_id]
            record_log(type='likeit', rec_id=rec_id, score=score,
                       context=context, prompt=prompt, response=response, data=data)
        self.slots['rec_id'] = None

    def face(self, text):
        return f'{self.face_icon}:{text}'

    def prompt(self, input_text, **kwargs):
        if self.slots['rec_id'] is not None:
            self.likeit(self.slots['rec_id'], 0)
            self.face_icon = '@kogi_minus'
        else:
            self.face_icon = '@kogi_plus'
        # 将来は分類モデルに置き換える
        # if self.slots.get('prompt_type') != 'direct':
            # if 'code' in self.slots:
            #     kwargs['include_code'] = True
            # if 'eline' in self.slots:
            #     kwargs['include_eline'] = True
            # if len(input_text) < 7: #コードを直して
            #     kwargs['detailed'] = True
        return self.dialog_with_context(input_text)

    def no_response(self):
        return 'AIが反応しない..'

    def get_prompt(self, input_text):
        if self.slots.get('prompt_type', '') == 'direct':
            return input_text
        ss = []
        if 'role' in self.slots:
            role = self.slots['role']
            ss.append(f'あなたは{role}です。')
        if 'about_me' in self.slots:
            about_me = self.slots['about_me']
            ss.append(f'私は{about_me}です。')
        ss.append(f'{input_text}')
        ss.append(f'80文字以内で簡潔にお願いします。')
        if 'tone' in self.slots:
            tone = self.slots['tone']
            ss.append(tone)
        return '\n'.join(ss)

    def get_context(self):
        if self.slots.get('prompt_type', '') == 'direct':
            return ''
        ss=[]
        if 'code' in self.slots:
            ss.append('コード:')
            ss.append('```')
            ss.append(remove_comment(self.slots['code']))
            ss.append('```')
        if 'emsg' in self.slots:
            emsg = self.slots['emsg']
            ss.append(f'発生したエラー: {emsg}')
        if 'eline' in self.slots:
            eline = self.slots['eline']
            ss.append(f'エラーの発生した行: {eline}')
        if 'prev_prompt' in self.slots:
            ss.append(f'前のプロンプト:')
            ss.append(self.slots['prev_prompt']+'\n')
        if 'prev_response' in self.slots:
            ss.append(f'前の回答:')
            ss.append(self.slots['prev_response']+'\n')
        return '\n'.join(ss)

    def dialog_with_context(self, input_text):
        prompt = self.get_prompt(input_text)
        context = self.get_context()
        debug_print(prompt, context, self.slots)
        response, tokens = model_prompt(prompt, context=context)
        if response == '':
            return self.no_response()
        self.slots['prev_prompt'] = input_text
        self.slots['prev_response'] = response
        rec_id = record_log(type='prompt',
                            prompt_type=self.slots.get('prompt_type', 'auto'), 
                            difftime=self.difftime(),
                            input_text=input_text,
                            context=context, prompt=prompt, response=response, tokens=tokens)
        self.chats[rec_id] = (context, prompt, response,
                              ('dialog_request', input_text))
        self.slots['rec_id'] = rec_id
        return self.face(response), rec_id


_DefaultChatbot = ChatAI()


def set_chatbot(chatbot):
    global _DefaultChatbot
    _DefaultChatbot = chatbot


LIKEIT = [0, 0]


def start_dialog(bot, start='', height=None, placeholder='質問はこちらに'):
    height = kogi_get('height', height)
    target = display_dialog(start, height, placeholder)

    def display_user(doc):
        nonlocal target
        append_message(doc, target, mention='@you')

    def display_bot_single(doc):
        nonlocal target
        append_message(doc, target)

    def display_bot(doc):
        if isinstance(doc, list):
            for d in doc:
                display_bot_single(d)
        else:
            display_bot_single(doc)

    if google_colab:
        def ask(user_text):
            global LIKEIT
            nonlocal bot
            try:
                if isinstance(user_text, str):
                    user_text = user_text.strip()
                debug_print(user_text)
                display_user(user_text)
                doc, rec_id = bot.prompt(user_text)
                doc = Doc.md(doc)
                doc.add_likeit(
                    rec_id, like=f'👍{LIKEIT[1]}', dislike=f'👎{LIKEIT[0]}')
                display_bot(doc)
            except:
                traceback.print_exc()
                display_bot('@robot:バグで処理に失敗しました。ごめんなさい')

        def like(docid, score):
            global LIKEIT
            nonlocal bot
            try:
                # debug_print(docid, score)
                bot.likeit(docid, score if score > 0 else -1)
                LIKEIT[score] += 1
            except:
                traceback.print_exc()
                display_bot('@robot:バグで処理に失敗しました。ごめんなさい')

        google_colab.register_callback('notebook.ask', ask)
        google_colab.register_callback('notebook.like', like)
        # if start != '':
        #     ask(start)
    return target


def call_and_start_kogi(actions, code: str = None, context: dict = None):
    for user_text in actions:
        _DefaultChatbot.update(context)
        _DefaultChatbot.slots['code']=code
        code = remove_comment(code)
        doc, rec_id = _DefaultChatbot.prompt(user_text)
        doc = Doc.md(doc)
        doc.add_likeit(rec_id)
        start_dialog(_DefaultChatbot, doc)
        return


def error_message(record):
    doc = Doc()
    if 'emsg_rewritten' in record:
        doc.println(record['emsg_rewritten'], bold=True)
        doc.println(record['emsg'], color='#888888')
    else:
        doc.println(record['emsg'])
        doc.println(record['_epat'], color='#888888')
    # print(record)
    if '_stacks' in record:
        for stack in record['_stacks'][::-1]:  # 逆順に
            if '-packages' in stack['filename']:
                continue
            doc.append(stack['_doc'])
    else:
        doc.append(record['_doc'])
    doc.set_mention('@kogi_minus')
    return doc


_HIRA_PAT = re.compile('[あ-を]')


def is_direct_kogi_call(record):
    if record.get('lineno') == 1 and record.get('etype') == 'SyntaxError':
        return True
    if record.get('etype') == 'NameError':
        eparams = record['_eparams']
        return re.search(_HIRA_PAT, eparams[0])
    return False


def kogi_prompt(prompt):
    _DefaultChatbot.update({'prompt_type': 'direct'})
    doc, rec_id = _DefaultChatbot.prompt(prompt)
    doc = Doc.md(doc)
    doc.add_likeit(rec_id)
    start_dialog(_DefaultChatbot, doc)

def catch_and_start_kogi(exc_info=None, code: str = None, context: dict = None, exception=None, enable_dialog=True):
    if exc_info is None:
        exc_info = sys.exc_info()
    record = kogi_exc(code=code, exc_info=exc_info,
                      caught_ex=exception, translate=translate)
    debug_print(record)
    if is_direct_kogi_call(record):
        kogi_prompt(code)
        return

    record_log(type='error', **record)
    messages = error_message(record)
    if context:
        record.update(context)
    _DefaultChatbot.update(record)
    start_dialog(_DefaultChatbot, start=messages)



import requests
from requests.exceptions import Timeout
from bs4 import BeautifulSoup

SAMPLE = {}


def download_atcoder_sample(url):
    if '?' in url:
        url, _, _ = url.rpartition('?')
    _, _, problem_id = url.rpartition('/')
    if problem_id in SAMPLE:
        return SAMPLE[problem_id]
    # try:
    response = requests.get(url, timeout=(3.0, 7.5))
    if response.status_code != 200:
        return {'error': f'読み込めません {response.status_code} {url}'}
    response_text = response.text
    html = BeautifulSoup(response_text, "lxml")
    d = {}
    for a in html.find_all("section"):
        # print(a)
        if a.h3 and a.pre:
            key = a.h3.text.replace('\r\n', '\n')
            value = a.pre.text.replace('\r\n', '\n')
            d[key] = value
    data = {'problem_id': problem_id, 'url': url}
    testcases = []
    if '入力例 1' in d:
        testcases.append(dict(input=d['入力例 1'], output=d['出力例 1']))
    if '入力例 2' in d:
        testcases.append(dict(input=d['入力例 2'], output=d['出力例 2']))
    if '入力例 3' in d:
        testcases.append(dict(input=d['入力例 3'], output=d['出力例 3']))
    data['testcases'] = testcases
    SAMPLE[problem_id] = data
    return data


# https://zenn.dev/hellorusk/articles/copy-from-blog-20210211
DROPBOX = 'https://www.dropbox.com/sh/nx3tnilzqz7df8a/AABP2YNyT09nYDTLV7d9bPwka/ABC102/A/in/sample_01.txt'
HOST = 'https://www.dropbox.com/sh/nx3tnilzqz7df8a/AACe0splFqRfbrMiRu41zmLna'


def download_atcoder_problem(directive):
    ss = directive.split()
    try:
        data = download_atcoder_sample(ss[0])
        return data
    except Timeout:
        return dict(error='ネットワーク接続できません')


if __name__ == '__main__':
    download_atcoder_problem(
        'https://atcoder.jp/contests/abc204/tasks/abc204_a random00.txt')



_GLOBALS = {
    'height': 360,
    'textra': 'cb25461ac40e7a2dc0b2bc05d381995a',
    # 'model_key': 'rhOcswxkXzMbhlkKQJfytbfxAPVsblhRHX',
}


def kogi_defined(key):
    global _GLOBALS
    return key in _GLOBALS


def kogi_get(key, value=None):
    global _GLOBALS
    return _GLOBALS.get(key, value)


def globals_update(data: dict):
    global _GLOBALS
    _GLOBALS.update(data)


def is_debugging():
    global _GLOBALS
    return _GLOBALS.get('debug', False)



import socket
import time
import signal
import sys
import os
import warnings
import traceback

_PREV_TIME = time.time()

try:
    from flask import Flask, request, jsonify
except ModuleNotFoundError:
    os.system('pip install -q flask')
    from flask import Flask, request, jsonify

app = Flask(__name__)


def _check_modules():
    try:
        import sentencepiece
    except:
        os.system('pip install -q sentencepiece')
    try:
        import transformers
    except:
        os.system('pip install -q transformers')

# NMT


_MODEL_ID = None
_MODEL = None
_TOKENIZER = None
_DEVICE = None
_CACHE = {}


def load_model(model_id):
    global _MODEL_ID, _MODEL, _TOKENIZER, _DEVICE, _CACHE
    if _MODEL_ID == model_id:
        return

    _check_modules()

    import torch
    from transformers import AutoTokenizer, AutoModelForSeq2SeqLM

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        tokenizer = AutoTokenizer.from_pretrained(model_id, is_fast=False)
        model = AutoModelForSeq2SeqLM.from_pretrained(model_id)

    # https://github.com/huggingface/transformers/issues/2542
    model = torch.quantization.quantize_dynamic(
        model, {torch.nn.Bilinear},
        dtype=torch.qint8
    )

    device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
    if isinstance(device, str):
        device = torch.device(device)
    model.to(device)

    _MODEL_ID = model_id
    _MODEL = model
    _TOKENIZER = tokenizer
    _DEVICE = device
    _CACHE = {}


def generate_gready(s: str, max_length=128) -> str:
    global _MODEL, _TOKENIZER, _DEVICE, _CACHE
    if _MODEL is None:
        return '<status>Now Loading'
    if s in _CACHE:
        return _CACHE[s]
    input_ids = _TOKENIZER.encode_plus(
        s,
        add_special_tokens=True,
        max_length=max_length,
        padding="do_not_pad",
        truncation=True,
        return_tensors='pt').input_ids.to(_DEVICE)

    greedy_output = _MODEL.generate(input_ids, max_length=max_length)
    t = _TOKENIZER.decode(greedy_output[0], skip_special_tokens=True)
    _CACHE[s] = t
    return t


def generate_beam(s: str, max_length=128, beam: int = 1) -> str:
    global _MODEL, _TOKENIZER, _DEVICE, _CACHE
    if _MODEL is None:
        return ['<status>Now Loading'], [0]

    key = f'{beam}:{s}'
    if key in _CACHE:
        return _CACHE[key]
    input_ids = _TOKENIZER.encode_plus(
        s,
        add_special_tokens=True,
        max_length=max_length,
        padding="do_not_pad",
        truncation=True,
        return_tensors='pt').input_ids.to(_DEVICE)

    # beem_search
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        outputs = _MODEL.generate(
            input_ids,
            # max_length=max_length,
            return_dict_in_generate=True, output_scores=True,
            temperature=1.0,          # 生成にランダム性を入れる温度パラメータ
            diversity_penalty=1.0,    # 生成結果の多様性を生み出すためのペナルティ
            num_beams=beam,
            #            no_repeat_ngram_size=2,
            num_beam_groups=beam,
            num_return_sequences=beam,
            repetition_penalty=1.5,   # 同じ文の繰り返し（モード崩壊）へのペナルティ
            early_stopping=True
        )
        results = [_TOKENIZER.decode(out, skip_special_tokens=True)
                   for out in outputs.sequences]
        scores = [float(x) for x in outputs.sequences_scores]
        _CACHE[key] = (results, scores)
        return results, scores


@app.route('/predict', methods=['GET', 'POST'])
def predict():
    global _PREV_TIME
    _PREV_TIME = time.time()
    content = request.get_json()
    # print(content)
    text = content.get('inputs', '')
    max_length = int(content.get('max_length', 128))
    beam = int(content.get('beam', 1))
    try:
        if beam <= 1:
            output = generate_gready(text, max_length=max_length)
            return jsonify({'model_id': _MODEL_ID, 'generated_text': output})
        outputs, scores = generate_beam(text, max_length=max_length, beam=beam)
        return jsonify({
            'model_id': _MODEL_ID,
            'generated_text': output[0],
            'outputs': outputs, 'scores': scores,
        })
    except Exception as e:
        tb = traceback.format_exc()
        return jsonify({
            'model_id': _MODEL_ID, 'generated_text': f'<error>{tb}',
        })


def get_pid():
    try:
        with open('.PID1789') as f:
            return int(f.read().strip())
    except:
        return None


def check_port(port):
    host = "127.0.0.1"
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((host, port))
    sock.close()
    return result == 0


def save_pid():
    pid = get_pid()
    if pid and check_port(5000):
        print('ALREADY LOADED', pid)
        return False
    with open('.PID1789', 'w') as w:
        print(f'{os.getpid()}', file=w)
    return True


def shutdown_server(signum, frame):
    print('Shutdown server...')
    try:
        func = request.environ.get('werkzeug.server.shutdown')
        if func:
            func()
    finally:
        pid = get_pid()
        if pid and pid == os.getpid():
            os.remove('.PID1789')


@app.route('/shutdown', methods=['GET'])
def shutdown():
    shutdown_server(None, None)
    return 'Server shutting down...'


signal.signal(signal.SIGTERM, shutdown_server)


def check_access(signum, frame):
    current_time = time.time()
    if current_time - _PREV_TIME > 600*6:
        shutdown_server()


signal.signal(signal.SIGALRM, check_access)
signal.setitimer(signal.ITIMER_REAL, 600, 600)  # 10分

if __name__ == '__main__':
    print('PID', get_pid(), os.getpid())
    if save_pid():
        load_model(sys.argv[1])
        app.run(debug=False, host='localhost')



from kogi.transform import get_kvars
from .common import Doc, debug_print
from .runner import task
from .diagnosis import IMPORT


def add_import(code):
    ss = []
    for key, snippet in IMPORT.items():
        if f'{key}.' in code:
            ss.append(snippet)
    if len(ss) == 0:
        return code
    return '\n'.join(ss) + '\n' + code


def _split_kw(s, kw):
    ss = []
    while True:
        left, sp, right = s.partion(kw)
        ss.append(left)
        if kw != sp:
            break
        ss.append(sp)
    return ss


def split_keywords(s, keywords):
    keywords.sort(key=lambda x: -len(x))
    ss = [s]
    for kw in keywords:
        ss2 = []
        for s in ss[::2]:
            if kw in s:
                ss2.extend(_split_kw(s, kw))
            else:
                ss2.append(s)
        ss = ss2
    return ss


@task('@translated_code')
def translated_code(bot, kwargs):
    doc = Doc()
    doc.println('こんなコードはいかが？')
    generated_code = add_import(kwargs['generated_text'])
    # doc.append(typewriter_doc(generated_code))
    code_doc = doc.new(style='@code')
    kvars = get_kvars(generated_code)
    for t in split_keywords(generated_code, kvars):
        if t in kvars:
            code_doc.append(t.replace('_', ''), style='@zen')
        else:
            code_doc.append(t)
    recid = bot.record("@translate_code",
                       kwargs['user_input'], kwargs['generated_text'])
    doc.add_likeit(recid, copy=generated_code)
    return doc


@task('@fix_code')
def fix_code(bot, kwargs):
    if 'eline' not in kwargs:
        debug_print(kwargs)
        return 'エラーが見つからないよ！'

    eline = kwargs['eline']
    tag, fixed = bot.generate(f'<コード修正>{eline}')
    recid = bot.record('@fix_code', eline, fixed)
    doc = Doc('コギー、がんばってみる')
    code_doc = doc.new(f'{eline}\n', style='@code')
    if tag != '<コード修正>' or eline == fixed:
        doc.println('ごめんね。なんかうまく直せないよ！')
    else:
        doc.println('どう？')
        code_doc = doc.new(f'{fixed}\n', style='@code')
        doc.add_likeit(recid, copy=fixed)
    return doc



from .globals import kogi_defined, kogi_get, globals_update
from .s3logging import kogi_print, print_nop, debug_print, record_log
from .textra import load_mt, translate
from .slack import load_slack, slack_send
from .chatgpt import set_openai, model_prompt
from .flaskapi import load_model, model_generate, check_awake, start_server


def kogi_set(**kwargs):
    globals_update(kwargs)
    if 'model_id' in kwargs:
        load_model(kwargs['model_id'])
    if 'mt_key' in kwargs:
        load_mt(kwargs['mt_key'])
    if 'slack_key' in kwargs:
        load_slack(kwargs['slack_key'])
    if 'textra_key' in kwargs:
        load_mt(kwargs['textra_key'])
    if 'openai_key' in kwargs:
        set_openai(kwargs['openai_key'])


def isEnglishDemo():
    return kogi_get('english_demo', False)



import re
import traceback
import warnings
from functools import wraps

# from kogi.logger import sync_lazy_loggger

from .chat import catch_and_start_kogi, call_and_start_kogi, record_log
from IPython.core.interactiveshell import InteractiveShell, ExecutionResult


RUN_CELL = InteractiveShell.run_cell
SHOW_TRACEBACK = InteractiveShell.showtraceback
SHOW_SYNTAXERROR = InteractiveShell.showsyntaxerror


KOGI_PAT = re.compile('#\\s*kogi\\s*(.*)')
HIRA_PAT = re.compile('[あ-を]')


def is_kogi_call(s):
    return re.search(HIRA_PAT, s)


def _find_action(text):
    return re.findall(KOGI_PAT, text)


def _call_kogi(code, actions):
    ss = []
    for action in actions:
        if is_kogi_call(action):
            ss.append(action)
    if len(ss) > 0:
        call_and_start_kogi(ss, code)
        return True
    return False


_DETECTOR = []
_RUNNER = {}


def kogi_register_hook(key, runner, detector):
    if key is not None and runner is not None:
        _RUNNER[key] = runner
    if detector is not None:
        _DETECTOR.append(detector)


def kogi_run_cell(ipy, raw_cell, kwargs):
    with warnings.catch_warnings():
        warnings.simplefilter('error', SyntaxWarning)
        result = None
        actions = _find_action(raw_cell)
        if len(actions) > 0:
            if _call_kogi(raw_cell, actions):
                return RUN_CELL(ipy, 'pass', **kwargs)
            for detector in _DETECTOR:
                key = detector(actions[0], raw_cell)
                if key in _RUNNER:
                    result = _RUNNER[key](
                        ipy, raw_cell, actions[0], catch_and_start_kogi)
                    if not isinstance(result, ExecutionResult):
                        result = RUN_CELL(ipy, 'pass', **kwargs)
                    return result
        if result is None:
            result = RUN_CELL(ipy, raw_cell, kwargs)
            if 'from google.colab.output import _js' not in raw_cell and raw_cell != "":
                record_log(type='run_cell', code=raw_cell)
        return result


def change_run_cell(func):
    @wraps(func)
    def run_cell(*args, **kwargs):
        try:
            # args[1] is raw_cell
            return kogi_run_cell(args[0], args[1], kwargs)
        except:
            traceback.print_exc()
        value = func(*args, **kwargs)
        return value
    return run_cell


def change_showtraceback(func):
    @wraps(func)
    def showtraceback(*args, **kwargs):
        try:
            ipyshell = args[0]
            code = ipyshell.user_global_ns['In'][-1]
            catch_and_start_kogi(code=code)
        except:
            traceback.print_exc()
    return showtraceback


def enable_kogi_hook():
    InteractiveShell.run_cell = change_run_cell(RUN_CELL)
    InteractiveShell.showtraceback = change_showtraceback(SHOW_TRACEBACK)
    InteractiveShell.showsyntaxerror = change_showtraceback(SHOW_SYNTAXERROR)


def disable_kogi_hook():
    InteractiveShell.run_cell = RUN_CELL
    InteractiveShell.showtraceback = SHOW_TRACEBACK
    InteractiveShell.showsyntaxerror = SHOW_SYNTAXERROR




from importlib import import_module
import re

from .common import debug_print, Doc, status_message
from .runner import model_parse, task, run_prompt
from kogi.liberr.rulebase import extract_params  # , expand_eparams
from kogi.data.diagnosis_ja import kwconv_diagnosis, reload

_CODE = re.compile(r'(`[^`]+`)')
_SPECIAL = re.compile(r'\<([^\>]+)\>')
_SVAR = re.compile(r'\<[A-Z]\>')
_OPTIONAL = re.compile(r'(\[[^\]]+\])')


def _extract_patterns(text, pat):
    return re.findall(pat, text)


def encode(text):
    codec = {}
    for code in _extract_patterns(text, _CODE):
        if re.search(_SVAR, code):
            continue
        key = f'@{id(code)}@'
        text = text.replace(code, key)
        codec[key] = code
    return text, codec


def decode(text, codec):
    for key, code in codec.items():
        text = text.replace(key, code)
    return text


def conv_nop(s):
    return s


def conv_unquote(s):
    if len(s) > 2 and s[0] in '"`\'' and s[-1] in '"`\'':
        return s[1:-1]
    return s


CONV_FUNC = {
    '': conv_nop,
    'unquote': conv_unquote,
}


def _replace_special_token(text, svar, kwargs):
    conv_fn = conv_unquote
    if '_' in svar:
        svar, _, funcname = svar.partition('_')
        if funcname not in CONV_FUNC:
            debug_print(f'{funcname} not in CONV_FUNC: {svar}_{funcname}')
            CONV_FUNC[funcname] = conv_unquote
        conv_fn = CONV_FUNC[funcname]
    if svar in kwargs:
        if svar < "D":
            replaced_text = conv_fn(str(kwargs[svar]))
        else:
            replaced_text = str(kwargs[svar])
    else:
        replaced_text = f'<{svar}>'  # そのまま
    return text.replace(f'<{svar}>', replaced_text)


def replace_special(text, kwargs):
    svars = _extract_patterns(text, _SPECIAL)
    for svar in svars:
        text = _replace_special_token(text, svar, kwargs)
        if f'<{svar}>' in text:
            return None
    return text


def select_option(option, kwargs):
    if '|' in option:
        for local_option in option.split('|'):
            local_option = replace_special(local_option, kwargs)
            if local_option is not None:
                return local_option
        return ''
    option = replace_special(option, kwargs)
    if option is None:
        return ''
    return option


def format_trago(text, kwargs):
    text, codec = encode(text)
    for option in _extract_patterns(text, _OPTIONAL):
        text = text.replace(option, select_option(option[1:-1], kwargs))
    text = decode(text, codec)
    return text

# UNDEFINED = collections.Counter()


def expand_keywords(text, d={}, bot=None, kwconv_fn=kwconv_diagnosis, UNDEFINED=None):
    keywords, kwargs = model_parse(text, d)
    if '_eparams' in kwargs:
        for X, val in zip('ABC', kwargs['_eparams']):
            kwargs[X] = val
    doc = Doc()
    for keyword in keywords:
        msg = kwconv_fn(keyword)
        if msg is None or msg == '':
            if UNDEFINED is not None:
                UNDEFINED.update([keyword])
            continue
        prompt_doc = None
        if '@' in msg:
            msg, _, cmd = msg.rpartition('@')
            cmd = f'@{cmd.strip()}'
            msg = msg.strip()
            prompt_doc = run_prompt(bot, cmd, kwargs)
        msg_doc = Doc.md(format_trago(msg, kwargs))
        doc.append(msg_doc)
        if prompt_doc:
            doc.append(prompt_doc)
    return doc


def convert_error_diagnosis(d, bot=None, UNDEFINED=None):
    etype, epat, eparams = extract_params(d['emsg'], maxlen=None)
    d['_etype'] = etype
    d['_epat'] = epat
    d['_eparams'] = eparams
    doc = expand_keywords(d['hint'], d, bot=bot, UNDEFINED=UNDEFINED)
    d2 = dict(
        eline=d['eline'],
        emsg=d['emsg'],
        epat=epat,
        eparams=eparams,
        hint=d['hint'],
        desc=str(doc).replace('\n', '')
    )
    return d2


@task('@root_cause_analysis @diagnosis @error')
def error_classfy(bot, kwargs):
    if 'emsg' not in kwargs or 'eline' not in kwargs:
        debug_print(kwargs)
        return 'エラーが見つからないよ！'
    emsg = kwargs['emsg']
    eline = kwargs['eline']
    input_text = f'<エラー分類>{eline}<sep>{emsg}'
    tag, fixed = bot.generate(input_text)
    if tag == '<status>':
        return status_message(fixed)
    if tag != '<エラー分類>':
        return 'うまく分析できないよ。ごめんね。'
    doc = expand_keywords(fixed, dict(kwargs), bot=bot)
    rec_id = bot.record('@error', input_text, fixed)
    doc.add_likeit(rec_id)
    return doc


IMPORT = {
    'math': 'import math',
    'os': 'import os',
    'sys': 'import sys',
    'random': 'import random',
    'datetime': 'import datetime',
    'collections': 'import collections',
    'builtins': 'import builtins',
    'copy': 'import copy',
    'np': 'import numpy as np',
    'pd': 'import pandas as pd',
    'plt': 'import matplotlib.pyplot as plt',
    'sns': 'import seaborn as sns',
    'scipy.stats': 'import scipy.stats',
}

FROM_IMPORT = {
    'DecisionTreeRegression': 'from sklearn.tree import DecisionTreeRegression',
    'DecisionTreeClassifier': 'from sklearn.tree import DecisionTreeClassifier',
}

MODULE_LIST = [
    'math',
    'collections',
    'sklearn.model_selection',
    'sklearn.metrics',
    'sklearn.metrics.pairwise',
    'sklearn.decomposition',
    'sklearn.linear_model',
    'sklearn.ensemble',
    'pytorch',
]


def init_module():
    for m in MODULE_LIST:
        try:
            for f in dir(import_module(m)):
                if not f.startswith('_'):
                    FROM_IMPORT.setdefault(f, f'from {m} import {f}')
        except ModuleNotFoundError:
            pass


init_module()


@task('@check_import')
def check_import(bot, kwargs):
    if 'A' not in kwargs:
        return None
    x = kwargs['A']
    if x in IMPORT:
        doc = Doc()
        doc.println('先に、次のモジュールをインポートを実行しておきましょう')
        doc.append(Doc(IMPORT[x], style='@code'))
        return doc
    if x in FROM_IMPORT:
        doc = Doc()
        doc.println('次の関数をインポートしてみましょう')
        doc.append(Doc(FROM_IMPORT[x], style='@code'))
        return doc
    return f'bot:「{x}をインポートするには？」'


@task('@check_zen')
def check_zen(bot, kwargs):
    if 'code' not in kwargs:
        return None
    code = kwargs['code']
    doc = Doc()
    doc.println('全角文字を赤く強調してみるよ')
    code_doc = doc.new(style='@code')
    for c in code:
        if ord(c) > 127:
            code_doc.append(c, style='@zen')
        else:
            code_doc.append(c)
    doc.println('コードの中では全角文字は使えないので直してね')
    return doc


@task('@xcopy')
def xcopy(args, kwargs):
    return '@ta:コピペは勉強にならないよ！'


@task('@xcall')
def xcall(bot, kwargs):
    return '先生は忙しいから、まずはTAさんに質問しましょう'




import os
import warnings
import requests
from .s3logging import kogi_print, print_nop
from .__async__ import is_loading, async_download

# NMT

_MODEL_ID = None
_MODEL = None
_TOKENIZER = None

_DEVICE = None


def _check_modules():
    try:
        import sentencepiece
    except:
        kogi_print('Installing sentencepiece')
        os.system('pip install -q sentencepiece')
    try:
        import transformers
    except:
        kogi_print('Installing transformers')
        os.system('pip install -q transformers')


_ASYNC_MODEL_ID = None


def load_model(model_id, qint8=True, device=None, async_downloading=False):
    global _MODEL_ID, _ASYNC_MODEL_ID, _MODEL, _TOKENIZER, _DEVICE
    if _MODEL_ID == model_id:
        return
    if async_downloading:
        _ASYNC_MODEL_ID = model_id
        async_download(model_id)
        return

    _ASYNC_MODEL_ID = None
    _check_modules()

    import torch
    from transformers import AutoTokenizer, AutoModelForSeq2SeqLM

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        tokenizer = AutoTokenizer.from_pretrained(model_id, is_fast=False)
        model = AutoModelForSeq2SeqLM.from_pretrained(model_id)

    if qint8:
        model = torch.quantization.quantize_dynamic(
            model, {torch.nn.Linear},
            dtype=torch.qint8
        )

    if device is None:
        device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
    if isinstance(device, str):
        device = torch.device(device)
    model.to(device)

    _MODEL_ID = model_id
    _MODEL = model
    _TOKENIZER = tokenizer
    _DEVICE = device


def generate_gready(s: str, max_length=128, print=print) -> str:
    global _MODEL, _TOKENIZER, _DEVICE
    input_ids = _TOKENIZER.encode_plus(
        s,
        add_special_tokens=True,
        max_length=max_length,
        padding="do_not_pad",
        truncation=True,
        return_tensors='pt').input_ids.to(_DEVICE)

    greedy_output = _MODEL.generate(input_ids, max_length=max_length)
    t = _TOKENIZER.decode(greedy_output[0], skip_special_tokens=True)
    #kogi_log(type='nmt', mode_id=_MODEL_ID, input=s, output=t)
    return t


def generate_beam(s: str, beam: int, max_length=12, print=print_nop) -> str:
    global _MODEL, _TOKENIZER, _DEVICE
    input_ids = _TOKENIZER.encode_plus(
        s,
        add_special_tokens=True,
        max_length=max_length,
        padding="do_not_pad",
        truncation=True,
        return_tensors='pt').input_ids.to(_DEVICE)

    # beem_search
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        outputs = _MODEL.generate(
            input_ids,
            # max_length=max_length,
            return_dict_in_generate=True, output_scores=True,
            temperature=1.0,          # 生成にランダム性を入れる温度パラメータ
            diversity_penalty=1.0,    # 生成結果の多様性を生み出すためのペナルティ
            num_beams=beam,
            #            no_repeat_ngram_size=2,
            num_beam_groups=beam,
            num_return_sequences=beam,
            repetition_penalty=1.5,   # 同じ文の繰り返し（モード崩壊）へのペナルティ
            early_stopping=True
        )
        results = [_TOKENIZER.decode(out, skip_special_tokens=True)
                   for out in outputs.sequences]
        scores = [float(x) for x in outputs.sequences_scores]
        return results, scores


API_URL = "https://api-inference.huggingface.co/models/kkuramitsu/kogi-mt5-test"
API_CACHE = {}


def generate_api(text, model_key, print=print_nop):
    global API_CACHE
    if len(text) > 120:
        return 'ぐるるるる\n（入力が長すぎます）'
    if text in API_CACHE:
        return API_CACHE[text]
    payload = {"inputs": text}
    headers = {"Authorization": f"Bearer {model_key}"}
    response = requests.post(API_URL, headers=headers, json=payload)
    output = response.json()
    # print(text, type(output), output)
    if isinstance(output, (list, tuple)):
        output = output[0]
    if 'generated_text' in output:
        result = output['generated_text']
        API_CACHE[text] = result
    return 'ねむねむ。まだ、起きられない！\n（しばらく待ってからもう一度試してください）'


def model_generate(text, beam=1, max_length=128, print=print_nop) -> str:
    global _MODEL, _ASYNC_MODEL_ID
    if _MODEL is None and _ASYNC_MODEL_ID:
        loading = is_loading()
        if loading:
            return f'<loading>{loading}をダウンロード中. しばらくお待ちください'
        load_model(_ASYNC_MODEL_ID)
    if _MODEL is not None:
        if beam > 1:
            return generate_beam(text, beam, max_length, print)
        return generate_gready(text, max_length, print)
    # if kogi_defined('model_key'):
    #     model_key = kogi_get('model_key')
    #     model_key = f'hf_{model_key}'
    #     return generate_api(text, model_key, print)
    return None



from IPython.display import display
from ipywidgets import Text, HTML
from .content import CSS
from .dialog import htmlfy_bot, htmlfy_user, Conversation


def htmlfy_dialog(chat, records, text=None):
    ss = [CSS('dialog.css')]
    ss.append('<div class="box2">')
    if text:
        records.append(text)
    ss.extend(records)
    ss.append('</div>')
    return ''.join(ss)

def display_dialog(chat, start=None, placeholder='質問はこちらに'):
    records=[]
    if start:
        start = htmlfy_bot(chat, start)
    html=HTML(value=htmlfy_dialog(chat, records, start))
    text = Text(value='',
                placeholder=placeholder,
                description='文字:',
                disabled=False
    )
    
    def dialog_bot(bot_text):
        nonlocal chat, records
        records.append(htmlfy_bot(chat, bot_text))
        html.value=htmlfy_dialog(chat, records)

    def dialog_user(user_text):
        nonlocal chat, records
        records.append(htmlfy_user(chat, user_text))
        html.value=htmlfy_dialog(chat, records)

    def update(submit):
        print(submit.value)
        user_text = str(submit.value).strip()
        dialog_user(user_text)
        bot_text = chat.ask(user_text)
        dialog_bot(bot_text)
        text.value=''

    text.on_submit(update)
    display(html)
    display(text)
    return dialog_bot, dialog_user




from .hook import enable_kogi_hook, disable_kogi_hook
from .service import kogi_set, debug_print
from .ui import kogi_print
import kogi.problem

set = kogi_set
print = kogi_print

enable = enable_kogi_hook
disable = disable_kogi_hook

try:
    enable_kogi_hook()
except:
    pass


import openai

model_cache = {}


def set_openai(api_key: str):
    openai.api_key = api_key


def model_chat(messages: list):
    try:
        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=messages,
        )
        used_tokens = response["usage"]["total_tokens"]
        res = response.choices[0]["message"]["content"].strip()
        return res, used_tokens
    except openai.error.AuthenticationError as e:
        return '', 0

def model_prompt(prompt, context='', **kwargs):
    global model_cache
    key = f'{context}{prompt}'
    if key in model_cache:
        return model_cache[key], 0
    try:
        response = openai.ChatCompletion.create(
            model="gpt-3.5-turbo",
            messages=[
                {"role": "system", "content": context},
                {"role": "user", f"content": f"{prompt}"},
            ],
        )
        used_tokens = response["usage"]["total_tokens"]
        res = response.choices[0]["message"]["content"].strip()
        model_cache[key] = res
        return res, used_tokens
    except openai.error.AuthenticationError as e:
        print("""
import kogi
kogi.set(openai_key="") # 自分のOpenAI キーを設定してください。""")
        return '', 0
    except openai.error.ServiceUnavailableError as e:
        print(e)
        return '', 0
    except Exception as e:
        print(e)
        return '', 0


import os
import string

from ._extract_emsg import extract_emsg, replace_eparams

try:
    import pegtree as pg
except ModuleNotFoundError:
    os.system('pip install pegtree')
    import pegtree as pg


_PEG = '''

Start = { 
    (Param /  { (!Param .)* } )*
}

Param = 
    / LongQuote / Quote 
    / Data / FuncName / CamelName / MaybeName / CellName/ VarName 
    / ClassName / PathName / UName
    / Float / Int / Hex

LongQuote =
    / {'\\'(' (!')\\'' . )+ ')\\'' #Quote}  // '(...)'
    / {'\\'<' (!'>\\'' . )+ '>\\'' #Quote}  // '<...>'
    / {'\\'[' ('\\\\' '\\'' / !']\\'' . )+ ']\\'' #Quote}   // '[  ]'

Quote =
    / SingleQuote
    / BackQuote
    / DoubleQuote

SingleQuote = { '\\'' (!'\\'' .)* '\\'' #Quote }
BackQuote = { '`' (!'`' .)* '`' #Quote }
DoubleQuote = { '"' (!'"' .)* '"' #Quote }

Data = Set / Tuple
Set = { '{' ( Data / !'}' . )* '}' #Set }
Tuple = { '[' ( Data / !']' . )* ']' #Tuple }

NAME = [A-Za-z_] [A-Za-z_.0-9]*
CAMEL = [A-Z]+ [a-z_0-9]+ [A-Z] NAME

FuncName = { NAME &('(x)' / '()') #FuncName }
CellName = { '%' '%'? NAME  #CellName }
CamelName = { CAMEL #CamelName }
VarName = { NAME ('\\'' NAME)? }  // can't
ClassName = { '<' [A-Za-z] (!'>' .)* '>' #ClassName }
PathName = 
    / { '(/' (!')' .)+ ')' #Path }
    / { '/usr' (![ ,] .)+ #Path}
MaybeName = 
    / { NAME &(' object' !NAME) #Maybe }
    / { NAME &(' instance' !NAME) #Maybe }
    / { NAME &(' expected' !NAME) #Maybe }

TYPENAME =
    / 'list' !NAME
    / 'tuple' !NAME
    / 'int' !NAME
    / 'float' !NAME
    / 'str' !NAME
    / 'deque' !NAME

Float = { '-'? [0-9]* '.' [0-9]+ #Number }
Int = { '-'? [0-9]+ ('.py')? ![A-Za-z] #Int }
Hex = { '0x' [0-9A-Fa-f]+ #Hex }

UName = { U (!END .)* #UName }
END = [ (),]

U = [ぁ-んァ-ヶ㐀-䶵一-龠々〇〻ー]

// python3 extract_emsg.py runtime_error-2022-*.jsonl syntax_error-2022-0*.jsonl kogi_chat-2022-08.jsonl kogi_chat-2022-09.jsonl kogi_chat-2022-10.jsonl exception_hook-2022-0*.jsonl undefined_error-2022-04.jsonl undefined_error-2022-05.jsonl unknown_error-2022-07.jsonl unknown_error-2022-08.jsonl unknown_emsg-2022-06.jsonl unknown_emsg-2022-07.jsonl unknown_emsg-2022-08.jsonl > ../emsg_debug.jsonl
'''

_parser = pg.generate(pg.grammar(_PEG))
_IDX = string.ascii_uppercase


def extract_params(emsg, maybe=True, maxlen=120):
    if '\n' in emsg:
        emsg = emsg.split('\n')[0]
    etype, _, emsg = emsg.partition(': ')
    tree = _parser(emsg)
    ss = []
    params = []
    for t in tree:
        s = str(t)
        if t == '':
            ss.append(s)
            continue
        if t == 'Maybe' and not maybe:
            ss.append(s)
            continue
        idx = _IDX[len(params) % 26]
        ss.append(f'<{idx}>')
        params.append(s)
    if maxlen:
        body = ''.join(ss)[:maxlen]
    else:
        body = ''.join(ss)
    return etype, body, params

# ルールベース


_EMSG_RULES = {}


def _abspath(file):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), file)


def load_local_rule(file='emsg_rules.tsv'):
    if not os.path.exists(file):
        file = _abspath(file)

    with open(file) as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            sentence = line.strip().split('\t')
            if len(sentence) == 2:
                _EMSG_RULES[sentence[0]] = sentence[1]


load_local_rule('emsg_rules.tsv')


def last_var(epat):
    for v in ('<D>', '<C>', '<B>', '<A>'):
        if v in epat:
            return v
    return None


def find_rule(epat):
    if epat in _EMSG_RULES:
        return _EMSG_RULES[epat]
    var = last_var(epat)
    if var is not None:
        epat, _, _ = epat.rpartition(var)
        #print('@', epat)
        for key, item in _EMSG_RULES.items():
            if key.startswith(epat):
                return item
    return None


UNQUOTE_FORMAT = '{}'


def _unquote(s):
    if s[0] == s[-1] and s[0] == "'" or s[0] == '`':
        s2 = s[1:-1]
        for c in s2:
            if ord(c) > 127 or not c.isalnum() and c != '_':
                return s
        return UNQUOTE_FORMAT.format(s2)
    return UNQUOTE_FORMAT.format(s)


def replace_eparams(msg, eparams):
    t = msg
    for X, val in zip(string.ascii_uppercase, eparams):
        t = t.replace(f'<{X}>', _unquote(val))
    return t


def _dequote(s):
    if len(s) > 2 and s[0] in '"`\'' and s[-1] in '"`\'':
        return s[1:-1]
    return s


def expand_eparams(record):
    if '_eparams' not in record:
        return
    for X, val in zip(string.ascii_uppercase, record['_eparams']):
        record[f'{X}_'] = _dequote(val)
        record[f'{X}'] = val


def rewrite_emsg(record, translate=None):
    emsg = record['emsg']
    etype, epat, eparams = extract_params(emsg, maxlen=None)
    record['_epat'] = epat
    record['_eparams'] = eparams
    translated = find_rule(f'{etype}: {epat}')
    if translated is None:
        if not translate:
            return None
        translated = translate(epat, lang='en_ja')
    if translated:
        translated = replace_eparams(translated, eparams)
        record['emsg_rewritten'] = translated
        return translated
    return None



import re

# mini md

_CODE = re.compile(r'(`[^`]+`)')
_BOLD = re.compile(r'(__[^_]+__)')


def replace_pre(s):
    while '```' in s:
        s = s.replace('```', '<pre>', 1).replace('```', '</pre>', 1)
    return s


def encode_md(s):
    s = s.replace('<', '&lt;').replace('>', '&gt;').replace('\n', '<br>')
    s = replace_pre(s)
    for t in re.findall(_CODE, s):
        t2 = f'<code>{t[1:-1]}</code>'
        s = s.replace(t, t2)
    for t in re.findall(_BOLD, s):
        t2 = f'<b>{t[2:-2]}</b>'
        s = s.replace(t, t2)
    return s


def encode_md_text(s):
    for t in re.findall(_CODE, s):
        t2 = t[1:-1]
        s = s.replace(t, t2)
    for t in re.findall(_BOLD, s):
        t2 = t[2:-2]
        s = s.replace(t, t2)
    return s


# try:
#     import markdown

#     def encode_md(s):
#         return markdown.markdown(s)
# except ModuleNotFoundError:
#     pass


TERM = {
    'code': '\033[35m{}\033[0m',
    'glay': '\033[07m{}\033[0m',
    'red': '\033[31m{}\033[0m',
    'green': '\033[32m{}\033[0m',
    'yellow': '\033[33m{}\033[0m',
    'blue': '\033[34m{}\033[0m',
    'magenta': '\033[35m{}\033[0m',
    'cyan': '\033[36m{}\033[0m',
}


def _term_div_color(color=None, background=None, bold=False):
    div = '{}'
    if color and color in TERM:
        div = TERM[color].format(div)
    if bold:
        div = '\033[01m{}\033[0m'.format(div)
    return div


def encode_md_term(s):
    for t in re.findall(_CODE, s):
        t2 = _term_div_color('code').format(t[1:-1])
        s = s.replace(t, t2)
    for t in re.findall(_BOLD, s):
        t2 = _term_div_color('bold').format(t[2:-2])
        s = s.replace(t, t2)
    return s


def _html_div_color(color=None, background=None, bold=False):
    div = '{}'
    if bold:
        div = f'<b>{div}</b>'
    if color and background:
        div = f'<span style="color: {color}; background: {background};">{div}</span>'
    elif color:
        div = f'<span style="color: {color};">{div}</span>'
    elif background:
        div = f'<span style="background: {background};">{div}</span>'
    return div


def _term(x):
    return x.term() if hasattr(x, 'term') else str(x)


_DEFAULT_STYLE = ('{}', '{}')

_STYLE_MAP = {
    '@pre': ('<pre>{}</pre>', '{}'),
    '@code': ('<pre class="code">{}</pre>', '{}'),
    '@zen': ('<span class="zen">{}</span>', '{}'),
}


def _get_style(style=None):
    if isinstance(style, str) and style.startswith('<'):
        return style, '{}'
    return _STYLE_MAP.get(style, _DEFAULT_STYLE)


def _tohtml(text):
    if '</' in text or '<br>' in text:
        return text.replace('</>', '')
    return text.replace('<', '&lt;').replace('>', '&gt;').replace('\n', '<br>')


def _html(x):
    return x._repr_html_() if hasattr(x, '_repr_html_') else _tohtml(str(x))


_FRAMEID = 1000


def frameid(frmid=None):
    global _FRAMEID
    if frmid:
        return frmid
    _FRAMEID += 1
    return _FRAMEID


class Doc(object):
    def __init__(self, doc=None, style=None):
        self.texts = []
        self.htmls = []
        self.terms = []
        self.style_format = _get_style(style)
        if doc:
            self.append(doc)

    def new(self, doc=None, style=None):
        doc = Doc(doc=doc, style=style)
        self.append(doc)
        return doc

    def __str__(self):
        content = ''.join(str(x) for x in self.texts)
        return content

    def term(self):
        content = ''.join(_term(x) for x in self.terms)
        return self.style_format[1].format(content)

    def __repr__(self):
        return self.term()

    def _repr_html_(self):
        content = ''.join(_html(x) for x in self.htmls)
        return self.style_format[0].format(content)

    def append(self, doc, style=None):
        if style is not None:
            doc = Doc(doc, style)
        if isinstance(doc, Doc):
            self.texts.append(doc)
            self.terms.append(doc)
            self.htmls.append(doc)
        elif doc is not None:
            self.texts.append(str(doc))
            self.terms.append(str(doc))
            self.htmls.append(_html(doc))

    def print(self, doc=None, style=None, color=None, background=None, bold=None):
        if style is not None:
            self.append(doc, style=style)
            return
        if isinstance(doc, Doc):
            self.texts.append(doc)
            self.terms.append(doc)
            self.htmls.append(doc)
        if isinstance(doc, str):
            doc = str(doc)
            self.texts.append(doc)
            div = _term_div_color(
                color=color, background=background, bold=bold)
            self.terms.append(div.format(doc))
            div = _html_div_color(
                color=color, background=background, bold=bold)
            self.htmls.append(div.format(_tohtml(doc)))

    def println(self, doc=None, style=None, color=None, background=None, bold=None):
        if doc:
            self.print(doc, style=style, color=color,
                       background=background, bold=bold)
        self.texts.append('\n')
        self.terms.append('\n')
        self.htmls.append('<br>')

    # def add_likeit(self, recid, frmid=None, copy=None, like='👍', dislike='👎'):
    #     frmid = frameid(frmid)
    #     if copy:
    #         textarea = f'<textarea id="t{frmid}" style="display: none">{{}}</textarea>'
    #         self.htmls.append(Doc(f'</>{copy}', style=textarea))
    #         button = f'<button id="b{frmid}" class="likeit" onclick="copy({frmid});like({recid},1)">{{}}</button>'
    #         self.htmls.append(Doc(f'コピー({like})', style=button))
    #     else:
    #         button = f'<button class="likeit" onclick="like({recid},1)">{{}}</button>'
    #         self.htmls.append(Doc(like, style=button))
    #     button = f'<button class="likeit" onclick="like({recid},0)">{{}}</button>'
    #     self.htmls.append(Doc(dislike, style=button))

    def add_likeit(self, recid, frmid=None, copy=None, like='👍1', dislike='👎0'):
        frmid = frameid(frmid)
        button = f'''\
<span id="b{frmid}">
<button class="likeit" onclick="like({recid},1,'b{frmid}')">{like}</button>
<button class="likeit" onclick="like({recid},0,'b{frmid}')">{{}}</button>
</span>'''
        self.htmls.append(Doc(dislike, style=button))

    def add_button(self, cmd, message, frmid=None):
        frmid = frameid(frmid)
        cmd = f"'{cmd}'"
        button = f'<button id="b{frmid}" onclick="say({cmd},{frmid})">{{}}</button>'
        self.htmls.append(Doc(message, style=button))

    def set_mention(self, mention: str):
        if mention.startswith('@'):
            mention, _, text = mention.partition(':')
            self.mention = mention
            return text
        return mention

    def get_mention(self, default=None):
        if hasattr(self, 'mention'):
            return self.mention
        for d in self.htmls:
            if isinstance(d, Doc):
                mention = d.get_mention()
                if mention:
                    return mention
        return default

    def get_script(self):
        script = ''
        if hasattr(self, 'script'):
            script = self.script
        for d in self.htmls:
            if isinstance(d, Doc):
                s = d.get_script()
                if s != '':
                    script += s
        return script

    @classmethod
    def md(cls, s, style=None):
        doc = Doc(style=style)
        s = doc.set_mention(s)
        doc.htmls.append(encode_md(s))
        doc.terms.append(encode_md_term(s))
        doc.texts.append(encode_md_text(s))
        return doc

    @classmethod
    def HTML(cls, html, text=None, css=None, script=None):
        doc = Doc()
        frmid = frameid()
        html = html.replace('ZYX', str(frmid))
        if css:
            html = css+html
        doc.htmls.append(html)
        if text:
            doc.terms.append(encode_md_term(text))
            doc.texts.append(encode_md_text(text))
        if script:
            script = script.replace('ZYX', str(frmid))
            doc.script = script
        return doc



# from ._extract_emsg import extract_emsg, replace_eparams
# from ._translate import translate_emsg, load_emsg_dic
# from ._traceback import kogi_print_exc, kogi_register_repr
from ._traceback2 import kogi_exc



from .s3logging import kogi_print, print_nop

_SLACK = None


def load_slack(slack_id):
    global _SLACK
    try:
        from slackweb import Slack
    except ModuleNotFoundError:
        import os
        os.system('pip install slackweb')
        from slackweb import Slack
    url = f'https://hooks.slack.com/services/{slack_id}'
    try:
        _SLACK = Slack(url)
    except Exception as e:
        kogi_print('Slackに接続できませんでした.', e)


def slack_send(text):
    global _SLACK
    if _SLACK is None:
        kogi_print('slack_idをセットしてください')
        return
    try:
        _SLACK.notify(text=text)
    except Exception as e:
        kogi_print('Slack Error:', e)



import sys
import re
import linecache
from kogi.ui.render import Doc

from .extract_vars import extract_vars
from .rulebase import rewrite_emsg


def DEBUG(*args, **kw):
    print('DEBUG', *args, **kw)


def getline(filename, lines, lineno):
    if filename == '<string>' or filename == '<unknown>':
        if 0 <= lineno-1 < len(lines):
            return lines[lineno-1]
        return ''
    return linecache.getline(filename, lineno).rstrip()

# def repr_value(value):
#     typename = type(value).__name__
#     if typename in REPR_VALUE:
#         return REPR_VALUE[typename](value)
#     if isinstance(value, str):
#         s = repr(value)
#         if len(s) > 32:
#             s = s[:32] + dots
#         return red(s)
#     if isinstance(value, Number) or value is None:
#         return repr(value)
#     return cyan(f'({typename})')


# def repr_vars(vars, start=None, end=None):
#     ss = []
#     for key, value in list(vars.items())[start:end]:
#         if key.startswith('_'):
#             continue
#         value = repr_value(value)
#         if value is not None:
#             ss.append(f'{bold(key)}={value}')
#     return ' '.join(ss)

def format_stack(doc, filename, funcname, local_vars, exprs=None, n_args=0):
    if filename.startswith('<ipython-input-'):
        t = funcname.split('-')
        if len(t) > 2:
            filename = f'[{t[2]}]'
    if '/ipykernel_' in filename:
        filename = ''
        if funcname.startswith('<'):
            funcname = ''
    if funcname != '':
        doc.print(funcname, color='blue', bold=True)
        doc.println(f' "{filename}"', color='glay')
    # arguments = repr_vars(local_vars, 0, n_args)
    # if len(arguments) > 2:
    #     arguments = f'({arguments})'
    # locals = repr_vars(filter_expressions(local_vars, exprs), n_args)
    # if '/ipykernel_' in filename:
    #     print(f'{bold(funcname)}{arguments}')
    # elif filename.endswith('.py'):
    #     print(f'"{glay(filename)}" {bold(funcname)}{arguments}')
    #     return
    # else:
    #     print(f'"{glay(filename)}" {bold(funcname)}{arguments}')
    # if len(locals) > 2:
    #     print(locals)


def format_arrow(doc, lineno, here=False):
    s = str(lineno)
    if here:
        arrow = '-' * max(5-len(s), 0) + '> '
    else:
        arrow = ' ' * max(5-len(s), 0) + '  '
    doc.print(arrow, color='red')
    doc.print(f'{s} ', color='green')


def format_linecode(doc, filename, lines, lineno):
    if lineno-2 > 0:
        format_arrow(doc, lineno-2)
        doc.println(getline(filename, lines, lineno-2))
    if lineno-1 > 0:
        format_arrow(doc, lineno-1)
        doc.println(getline(filename, lines, lineno-1))
    format_arrow(doc, lineno, here=True)
    doc.println(getline(filename, lines, lineno))
    # render_arrow(doc, lineno+1)
    # doc.println(getline(filename, lines, lineno+1))
    # render_arrow(doc, lineno+2)
    # doc.println(getline(filename, lines, lineno-2))


def format_offset(doc, lineno, offset):
    offset = max(0, offset-1)
    format_arrow(doc, lineno)
    doc.print(' ' * offset)
    doc.println('^^', color='red', bold=True)


def syntax_exc(code, caught_ex, record):
    filename = caught_ex.filename
    lines = code.splitlines()
    record['lineno'] = lineno = caught_ex.lineno
    record['eline'] = text = caught_ex.text
    record['offset'] = offset = caught_ex.offset
    doc = Doc(style='@pre')
    format_linecode(doc, filename, lines, lineno)
    format_offset(doc, lineno, offset)
    # print(dir(caught_ex))
    record['_doc'] = doc
    return record


def filter_expressions(vars, exprs=None):
    if exprs is not None:
        newvars = {}
        for key, value in vars.items():
            if key in exprs:
                newvars[key] = value
        return newvars
    return vars


_VARPAT = re.compile(r'([A-Za-z_][A-Za-z_]*)')


def find_var(parent, eline, locals={}):
    doc = Doc(style='@pre')
    dup = set()
    found = re.findall(_VARPAT, eline)
    if found:
        for name in found:
            if name in locals and name not in dup:
                dup.add(name)
                format_value(doc, locals[name], name)
    if len(dup) > 0:
        doc = Doc(doc, style='<details><summary>変数の値を確認する</summary>{}</details>')
        parent.append(doc)


def format_value(doc, value, name=None):
    if name:
        doc.print(f'   {name}', bold=True)
        doc.print(f' = ')
    doc.print(f'{type(value).__name__}型 ', color='green')
    if hasattr(value, 'shape'):
        doc.print(f'shape:{repr(value.shape)} ', color='green')
    elif hasattr(value, '__len__'):
        doc.print(f'len:{len(value)} ', color='green')
    dump = repr(value)
    if len(dump) > 20:
        dump = dump + '...'
    doc.println(dump)


def runtime_exc(code, tb, record):
    record['traceback_'] = tb
    exprs = None
    if code != '':
        exprs = extract_vars(code)
        record['exprs_in_code'] = exprs
        exprs = set(exprs)
    lines = code.splitlines()

    stacks = []
    eline = None

    while tb:
        filename = tb.tb_frame.f_code.co_filename
        funcname = tb.tb_frame.f_code.co_name
        lineno = tb.tb_lineno
        line = getline(filename, lines, tb.tb_lineno)
        n_args = tb.tb_frame.f_code.co_argcount
        local_vars = tb.tb_frame.f_locals
        stack = dict(
            filename=filename,
            funcname=funcname,
            n_args=n_args,
            local_vars=local_vars,
            lineno=lineno, line=line,
            line_in_code=line in code,
        )
        doc = Doc()
        format_stack(doc, filename, funcname, local_vars, exprs, n_args)
        pre = Doc(style='@pre')
        format_linecode(pre, filename, lines, lineno)
        doc.append(pre)
        find_var(doc, line, local_vars)
        stack['_doc'] = doc
        stacks.append(stack)
        if line in code:
            eline = line
        tb = tb.tb_next
    record['_stacks'] = stacks
    if eline:
        record['eline'] = eline


def kogi_exc(code='', exc_info=None, caught_ex=None, translate=None):
    if exc_info is None:
        etype, evalue, tb = sys.exc_info()
    else:
        etype, evalue, tb = exc_info
    if etype is None:
        return None
    record = dict(
        etype=f'{etype.__name__}',
        emsg=(f'{etype.__name__}: {evalue}').strip(),
        code=code,
    )
    if caught_ex is None and issubclass(etype, SyntaxError):
        try:
            raise
        except SyntaxError as e:
            caught_ex = e
    if isinstance(caught_ex, SyntaxError):
        syntax_exc(code, caught_ex, record)
    else:
        runtime_exc(code, tb, record)
    rewrite_emsg(record, translate=translate)
    return record


def print_record(record):
    r = Render()
    if 'emsg_rewritten' in record:
        r.println(record['emsg_rewritten'], bold=True)
        r.println(record['emsg'])
    else:
        r.println(record['emsg'])
    print(r.termtext())
    if '_stacks' in record:
        for stack in record['_stacks']:
            print_record(stack)
    elif '_term' in record:
        print(record['_term'])
    else:
        print(record['_text'])
    print(r.termtext())


def print_exc(code='', exc_info=None, caught_ex=None, translate=None):
    record = kogi_exc(code, exc_info, caught_ex, translate)
    print_record(record)



import traceback
from ._google import google_colab

from kogi.service import debug_print
from .message import display_dialog, append_message

# <div id="{target}" class="box" style="height: {height}px"></div>


def start_dialog(bot, start='', height=None, placeholder='質問はこちらに'):
    target = display_dialog(start, height, placeholder)

    def display_user(doc):
        nonlocal target
        append_message(doc, target, mention='@you')

    def display_bot_single(doc):
        nonlocal target
        append_message(doc, target)

    def display_bot(doc):
        if isinstance(doc, list):
            for d in doc:
                display_bot_single(d)
        else:
            display_bot_single(doc)

    if google_colab:
        def ask(user_text):
            nonlocal bot
            try:
                user_text = user_text.strip()
                debug_print(user_text)
                display_user(user_text)
                doc = bot.ask(user_text)
                display_bot(doc)
            except:
                traceback.print_exc()
                display_bot('@robot:バグで処理に失敗しました。ごめんなさい')

        def like(docid, score):
            nonlocal bot
            try:
                debug_print(docid, score)
                bot.log_likeit(docid, score)
            except:
                traceback.print_exc()
                display_bot('@robot:バグで処理に失敗しました。ごめんなさい')

        def say(prompt, text):
            nonlocal bot
            try:
                debug_print(text, prompt)
                display_user(text)
                doc = bot.exec(prompt)
                display_bot(doc)
            except:
                traceback.print_exc()
                display_bot('@robot:バグで処理に失敗しました。ごめんなさい')

        google_colab.register_callback('notebook.ask', ask)
        google_colab.register_callback('notebook.like', like)
        google_colab.register_callback('notebook.say', say)
    return target



import inspect
import pytz
import uuid
import json
# import traceback
# import signal
import requests
from datetime import datetime, timezone
from .globals import kogi_get, is_debugging


def kogi_print(*args, **kw):
    print('\033[34m[🐶]', *args, **kw)
    print('\033[0m', end='')


def debug_print(*args, **kw):
    if is_debugging():
        filename = inspect.currentframe().f_back.f_code.co_filename
        lineno = inspect.currentframe().f_back.f_lineno
        if '/kogi/' in filename:
            _, _, filename = filename.rpartition('/kogi/')
        loc = f'[🐝{filename}:{lineno}]'
        print('\033[35m' + loc, *args, **kw)
        print('\033[0m', end='')


def print_nop(*args, **kw):
    pass


SESSION = str(uuid.uuid1())
SEQ = 0
_LOG_BUFFERS = []


def _copylog(logdata):
    if isinstance(logdata, dict):
        copied = {}
        for key, value in logdata.items():
            if key.startswith('_') or key.endswith('_'):
                continue
            copied[key] = _copylog(value)
        return copied
    if isinstance(logdata, list) or isinstance(logdata, tuple):
        return [_copylog(x) for x in logdata]
    return logdata


def record_log(lazy=False, **kargs):
    global SEQ, _LOG_BUFFERS, epoch
    now = datetime.now(pytz.timezone('Asia/Tokyo'))
    date = now.isoformat(timespec='seconds')
    logdata = _copylog(dict(seq=SEQ, date=date, **kargs))
    if 'type' not in logdata:
        logdata['type'] = 'debug'
    if is_debugging():
        logdata['type_orig'] = logdata['type']
        logdata['type'] = 'debug'
    SEQ += 1
    if kogi_get('approved', False):
        _LOG_BUFFERS.append(logdata)
        send_log(not lazy)
    return SEQ-1


UID = 'unknown'
POINT = 'ixe8peqfii'
HOST2 = 'amazonaws'
KEY = 'OjwoF3m0l20OFidHsRea3ptuQRfQL10ahbEtLa'
prev_epoch = datetime.now().timestamp()


def send_log(right_now=True):
    global prev_epoch, _LOG_BUFFERS, POINT
    now = datetime.now().timestamp()
    delta = (now - prev_epoch)
    prev_epoch = now
    if len(_LOG_BUFFERS) > 0 and (right_now or delta > 30 or len(_LOG_BUFFERS) > 4):
        data = {
            "session": SESSION,
            "uname": kogi_get('uname', ''),
            "approved": kogi_get('approved', False),
            "logs": _LOG_BUFFERS,
        }
        url = f'https://{POINT}.execute-api.ap-northeast-1.{HOST2}.com/dev'
        headers = {'x-api-key': f'A{KEY}s'}
        r = requests.post(url, headers=headers, json=data)
        debug_print('logging', data)
        _LOG_BUFFERS.clear()
        if r.status_code != 200:
            debug_print(r.status_code)
            debug_print(r)





import IPython
from IPython.display import JSON
from ._google import google_colab
from .render import Doc

_CSS = """\
<style>
.lds-facebook {
  display: inline-block; position: relative; 
  width: 60px; height: 30px;
}
.lds-facebook div {
  display: inline-block; position: absolute;
  left: 8px; width: 16px; background: #ebb67f;
  animation: lds-facebook 1.2s cubic-bezier(0, 0.5, 0.5, 1) infinite;
}
.lds-facebook div:nth-child(1) {
    left: 8px; animation-delay: -0.24s;
}
.lds-facebook div:nth-child(2) {
    left: 32px; animation-delay: -0.12s;
}
.lds-facebook div:nth-child(3) {
    left: 56px; animation-delay: 0; 
}
@keyframes lds-facebook {
  0% { top: 8px; height: 32px; }
  50%, 100% { top: 24px; height: 16px; }
}
</style>
"""

_HTML = """\
<div id='XYZloading'>
<span id='XYZloading_msg'></span>
<div class="lds-facebook"><div></div><div></div><div></div></div>
</div>
"""

_JS = """\
<script>
const check = () => {
    (async function() {
    const result = await google.colab.kernel.invokeFunction('notebook.Waiting', [], {})
    const data = result.data['application/json'];
    if(!data.result) {
        const loading = document.getElementById('XYZloading');
        loading.innerText=data.message;
    }
    else {
        const loading = document.getElementById('XYZloading_msg');
        loading.innerText=data.message;
        setTimeout(check, 1000)
    }
    })();
}
check()
</script>
"""

c = 0


def check_ready_dummy():
    global c
    c += 1
    return c > 5


def wait_for_ready_doc(check_ready_fn=check_ready_dummy,
                       wait='コギーを起こしています。（まだ、眠っています。）',
                       ready='コギーが起きた！おはよ'):
    global c
    c = 0

    def waiting_callback():
        if check_ready_fn():
            return JSON({'result': False, 'message': ready})
        return JSON({'result': True, 'message': wait})
    if google_colab:
        google_colab.register_callback('notebook.Waiting', waiting_callback)
    return Doc.HTML(css=_CSS, html=_HTML+_JS, text=wait)



from kogi.service import kogi_get, debug_print
from .content import ICON, CSS, JS
from kogi.ui.render import Doc

from IPython.display import display, HTML
from ._google import google_colab

_ICON = {
    '@kogi': ('コギー', 'kogi_doya-fs8.png'),
    '@kogi_plus': ('コギー', 'kogi_doya-fs8.png'),
    '@kogi_minus': ('コギー', 'kogi_error-fs8.png'),
    '@robot': ('コギー', 'kogi_gaan-fs8.png'),
    '@ta': ('ももパン', 'ta-fs8.png'),
    '@you': ('あなた', 'girl-fs8.png'),
}

def get_icon(tag):
    return _ICON.get(tag, _ICON['@kogi'])


_BOT_HTML = '''\
<div class="sb-box">
<div class="icon-img icon-img-left"><img src="{icon}" width="60px"></div>
<div class="icon-name icon-name-left">{name}</div>
<div class="sb-side sb-side-left"><div class="sb-txt sb-txt-left">{content}</div></div>
</div>
'''

_USER_HTML = '''\
<div class="sb-box">
<div class="icon-img icon-img-right"><img src="{icon}" width="60px"></div>
<div class="icon-name icon-name-right">{name}</div>
<div class="sb-side sb-side-right"><div class="sb-txt sb-txt-right">{content}</div></div>
</div>
'''


def messagefy(doc, mention=None):
    if isinstance(doc, str):
        if doc.startswith('@'):
            mention, _, text = doc.partition(':')
            doc = Doc(text)
        else:
            doc = Doc(doc)
    if mention is None:
        mention = doc.get_mention('@kogi')
    name, icon = get_icon(mention)
    if mention == '@you':
        return _USER_HTML.format(
            icon=ICON(icon),
            name=kogi_get('uname', name),
            content=doc._repr_html_(),
        ), doc.get_script()
    return _BOT_HTML.format(
        icon=ICON(icon),
        name=name,
        content=doc._repr_html_(),
    ), doc.get_script()


_DIALOG_ID = 0


def replace_dialog_id(s):
    return s.replace('XYZ', str(_DIALOG_ID))


def display_dialog_css():
    display(HTML(CSS('dialog.css')))


def dialog_script():
    return JS('dialog.js')


def exec_js(script):
    if script != '':
        display(HTML(f'<script>\n{script}</script>'))


_DIALOG = '''\
<div id="dialogXYZ" class="box">{}</div>
'''

_DIALOG2 = '''\
<div id="dialogXYZ" class="box" style="height: {}px">{}</div>
'''

_TEXTAREA = '''\
<div style="text-align: right">
<textarea id="inputXYZ" placeholder="@placeholder@"></textarea>
<script>
let timeout = 3*60*1000;
var tm = setTimeout(()=>{document.getElementById("inputXYZ").remove();}, timeout);
document.getElementById("inputXYZ").addEventListener('keydown', (e) => {
    if (e.keyCode == 13) {
        const pane = document.getElementById("inputXYZ");
        clearTimeout(tm);
        tm = setTimeout(()=>{pane.remove();}, timeout*2);
        google.colab.kernel.invokeFunction('notebook.ask', [pane.value], {});
        pane.value = '';
    }
});
</script>
</div>
'''

_BLOCK = '''\
<div>
{}
{}
</div>
'''


def display_dialog(doc='', height=None, placeholder=None):
    global _DIALOG_ID
    _DIALOG_ID += 1
    display_dialog_css()
    if doc == '':
        html, script = '', ''
    else:
        html, script = messagefy(doc)
    if height:
        html = _DIALOG.format(html)
        ##html = _DIALOG2.format(height, html)
    else:
        html = _DIALOG.format(html)
    if placeholder:
        html = html+_TEXTAREA.replace('@placeholder@', placeholder)
    html = _BLOCK.format(dialog_script(), html)
    html = replace_dialog_id(html)
    script = replace_dialog_id(script)
    display(HTML(html))
    exec_js(script)
    return replace_dialog_id('#dialogXYZ')


APPEND_JS = '''\
<script>
var target = document.getElementById("dialogXYZ");
var content = `{html}`;
if(target !== undefined) {{
    target.insertAdjacentHTML('beforeend', content);
    target.scrollTop = target.scrollHeight;
}}
</script>
'''


def append_message(doc, target, mention=None):
    html, script = messagefy(doc, mention=mention)
    # if google_colab:
    #     with google_colab.redirect_to_element(target):
    #         display(HTML(replace_dialog_id(html)))
    # else:
    html = html.replace('\\', '\\\\')
    html = html.replace('`', '\\`')
    display(HTML(replace_dialog_id(APPEND_JS).format(html=html)))
    exec_js(replace_dialog_id(script))
    return target


def kogi_print(*args, **kwargs):
    height = kwargs.get('height', None)
    target = kwargs.get('target', None)
    placeholder = kwargs.get('placeholder', None)
    if len(args) > 0:
        if isinstance(args[0], Doc):
            if target:
                return append_message(args[0], target)
            else:
                return display_dialog(args[0], placeholder=placeholder, height=height)
    sep = kwargs.get('sep', ' ')
    text = sep.join(str(s) for s in args)
    if target:
        return append_message(text, target)
    else:
        return display_dialog(text, placeholder=placeholder, height=height)



import os

_EDESC = {}


def _abspath(file):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), file)


def load_error(file='error.tsv'):
    if not os.path.exists(file):
        file = _abspath(file)

    with open(file) as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            kw, _, edesc = line.partition(': ')
            if edesc != '':
                kw = kw.strip()
                edesc = edesc.strip()
                _EDESC[kw] = edesc
                #print(kw, edesc)


load_error('error.txt')


def get_error_desc(kw):
    if kw in _EDESC:
        return _EDESC[kw]
    return ''



import sys
import numbers
import string
import os
import pandas as pd
from IPython import get_ipython
from kogi.service import debug_print, model_generate, kogi_set
from .OLDconversation import ConversationAI, set_chatbot
try:
    import pegtree as pg
except ModuleNotFoundError:
    os.system('pip install pegtree')
    import pegtree as pg

kogi_set(
    model_id='NaoS2/multi-kogi',
    debug=True,
)


def extract_tag(text):
    if text.startswith('<'):
        tag, end_tag, text = text.partition('>')
        return tag+end_tag, text
    return '', text


# ユーザの入力→モデルの入力
# 実行時型取得
KOGI_TYPEMAP = {
    'bool': '_結果_',
    'int': '_整数_',
    'float': '_数値_',
    'tuple': '_タプル_',
    'list': '_リスト_',
    'str': '_文字列_',
    'function': '_関数_',
    'ndarray': '_配列_',
    'DataFrame': '_データフレーム_',
}


def get_kogitype(value):
    py_type = type(value).__name__
    kg_type = KOGI_TYPEMAP.get(py_type, None)
    if kg_type is not None:
        return kg_type
    if isinstance(value, numbers.Number):
        return '_数値_'
    if getattr(value, '__iter__'):
        return '_イテラブル_'
    return f'_結果_'


possy = {'n': '_整数_', 'N': '_整数_', }  # みんなが使う変数名


def get_variable_type(name):
    shell = get_ipython()
    if name in shell.user_ns:
        value = shell.user_ns[name]
        return get_kogitype(value)
    else:  # 変数が未定義な場合
        return possy.get(name, '_結果_')


def eval_code_type(code):
    shell = get_ipython()
    try:
        value = eval(code, shell.user_ns)
        return get_kogitype(value)
    except:
        return '_結果_'


# pegtree
_PEG = '''

Start = { 
    (Chunk /  { (!Chunk .)* #Chunk } )*
}

Chunk = 
    / SingleQuote
    / DoubleQuote
    / BackQuoteCode
    / Number
    / Variable

SingleQuote = { '\\'' (!'\\'' .)* '\\'' #String } // ' ... '
DoubleQuote = { '"' (!'"' .)* '"' #String }  // " "
BackQuoteCode = '`' {  (!'`' .)*  #Code } '`' // `1+2`

Number = { '-'? [0-9]+ ('.' [0-9]+)? #Number } // 1.2
Variable = { [A-Za-z_0-9]+ #Name }  // a

'''

_parser = pg.generate(pg.grammar(_PEG))


def scan_dataframes():
    column_maps = {}
    dataframe_names = []
    shell = get_ipython()
    user_ns = shell.user_ns
    for name in user_ns:
        if name[0] == "_":  # 仮
            pass
        else:
            value = user_ns[name]
            if isinstance(value, pd.DataFrame):
                dataframe_names.append(name)
                for column in list(value.columns):
                    column_maps.setdefault(column, name)
    return dataframe_names, column_maps


def detect_string_type(s):
    content = s[1:-1]  # クオートをとる
    if content.endswith('.csv'):
        return '_CSVファイル_'
    if len(content) == 1:
        return '_文字_'
    return '_文字列_'


def append_map(maps, key, value):
    if key not in maps:
        maps[key] = []
    maps[key].append(value)

# parser


def parse(text):
    dataframe_names, column_maps = scan_dataframes()
    after_maps = {}
    tree = _parser(text)
    ss = []
    for t in tree:
        token = str(t)
        if t == 'Number':
            kgtype = '_数値_' if '.' in token else '_整数_'
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'String':
            kgtype = detect_string_type(token)
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'Name':
            kgtype = get_variable_type(token)
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'Code':
            kgtype = eval_code_type(token)
            append_map(after_maps, kgtype, token)
            ss.append(kgtype)
        elif t == 'Chunk':
            for column, dfname in column_maps.items():
                if column in token:
                    append_map(after_maps, '_列名_', f'"{column}"')
                    append_map(after_maps, '_データフレーム_', dfname)
                    append_map(after_maps, '_データ列_', f'{dfname}["{column}"]')
                    token = token.replace(column, '_列名_')
            ss.append(token)
    return ''.join(ss).replace('_', ''), after_maps


# モデル出力→ユーザへの出力
# 下線区切り
def get_words(text):
    words = []
    while True:
        if len(text) > 0:
            pos1 = text.find('_')
            pos2 = text.find('_', pos1+1)
            if pos1 == -1 or pos2 == -1:
                break
            else:
                word = text[pos1:pos2+1]
                words.append(word)
                text = text[pos2:]
        else:
            break
    return words


def make_output(text, dic):
    words = get_words(text)
    for word in words:
        if word in dic:
            new_word = dic[word].pop(0)
            text = text.replace(word, new_word, 1)
            dic[word].append(new_word)
    return text


# <tag>省略
def extract_tag(text):
    if text.startswith('<'):
        tag, end_tag, text = text.partition('>')
        return tag+end_tag, text
    return '', text


class PanAI(ConversationAI):

    def response(self, user_input):
        user_input, after_maps = parse(user_input)  # ユーザ入力→モデルの入力
        debug_print("user_input:"+user_input)
        response_text = model_generate(user_input)  # 予測
        if response_text is None:
            return 'ZZ.. zzz.. 眠む眠む..'
        debug_print("response_text:"+response_text)
        response_text = make_output(
            response_text, after_maps)  # モデルの出力→ユーザへの出力

        tag, text = extract_tag(response_text)
        return text


debug = False
set_chatbot(PanAI())



import os


def _abspath(file):
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), file)


def load_data(file, DATA, default_first=True):
    if not os.path.exists(file):
        file = _abspath(file)

    with open(file) as f:
        for line in f.readlines():
            if line.startswith('#'):
                continue
            kw, _, data = line.partition(': ')
            if data != '':
                kw = kw.strip()
                data = data.strip()
                if default_first:
                    DATA.setdefault(kw, data)
                else:
                    DATA[kw] = data



import os
from .common_loader import load_data

_KWDATA = {}
load_data('diagnosis_ja.txt', _KWDATA)


def kwconv_diagnosis(kw, default=''):
    if kw in _KWDATA:
        return _KWDATA[kw]
    return default


def reload(filename):
    global _KWDATA
    _KWDATA = {}
    load_data(filename, _KWDATA)



from .render import Doc

_CSS = """\
<style>
.var-highlight{
  color: #C0AD60;
}
.string-highlight{
  color: rgba(253, 149, 90, 0.8);
}
.typewriter{
    font-size: 2em;
    margin: 0;
    font-family: monotype;
    &:after{
      content: "|";
      animation: blink 500ms linear infinite alternate;
    }
}
@keyframes blink{
  0%{opacity: 0;}
  100%{opacity: 1;}
}
</style>
"""

_JS = """\
const typer = document.getElementById('typewriterZYX');
typewriter = setupTypewriter(typer);
typewriter.type();
"""

_HTML = """\
<pre id="typewriterZXY" class="typewriter">{}</pre>
"""


def typewriter_doc(text):
    return Doc.HTML(html=_HTML.format(text), text=text, css=_CSS, script=_JS)




from ._google import google_colab
# from ._rmt import TransformWeaver, rmt
from .content import ICON
from .message import kogi_print

if google_colab:
    from .dialog_colab import display_dialog
else:
    try:
        import ipywidgets
        from .dialog_ipywidgets import display_dialog
    except ModuleNotFoundError:
        from .dialog_colab import display_dialog



import traceback
from ._google import google_colab

from IPython.display import display, HTML, JSON
from kogi.service import model_generate, debug_print

RMT_CSS = '''
<style>
.parent {
    background-color: #edebeb;
    width: 100%;
    height: 150px;
}
textarea {
    width: 100%;
    box-sizing: border-box;
    /* ※これがないと横にはみ出る */
    height: 120px;
    font-size: large;
    outline: none;
    /* ※ブラウザが標準で付加する線を消したいとき */
    resize: none;
}
.outbox {
    /*padding: 0.2em 0.5em; margin: 2em 0; */
    color: #565656;
    background: #ffeaea;
    background-size: 150%;
    background-repeat: no-repeat;
    background-position: top right;
    background-color: rgba(255, 255, 255, 0.8);
    background-blend-mode: lighten;
    border: dashed 2px #ffc3c3;
    height: 120px;
    font-size: large;
    text-align: left;
}
.inbox {
    /*
    padding: 0.5em 1em;margin: 2em 0;
    */
    background: -webkit-repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px, #e9f4ff 3px, #e9f4ff 7px);
    background: repeating-linear-gradient(-45deg, #f0f8ff, #f0f8ff 3px, #e9f4ff 3px, #e9f4ff 7px);
}
.labelbox {
    position: relative;
    padding: 0.5em 0.7em;
    margin: 2em 0;
    background: #6f4b3e;
    color: white;
    font-weight: bold;
}
</style>
'''

RMT_HTML = '''
<div id="{id}" class="parent">
<div style="float: left; width: 48%; text-align: right;">
<label class="labelbox" for="input">{input_caption}</label>
<textarea id="input" class="inbox"></textarea>
</div>
<div style="float: left; width: 48%; text-align: right;">
<label class="labelbox" for="outout">{output_caption}</label>
<div id="output" class="outbox"></div>
</div>
</div>
'''

RMT_JS = '''
<script>
var timer = null;
var inputPane = document.getElementById('input');
inputPane.addEventListener('input', (e) => {
    var text = e.srcElement.value;
    if (timer !== null) {
        clearTimeout(timer);
    }
    timer = setTimeout(() => {
        timer = null;
        (async function () {
            const result = await google.colab.kernel.invokeFunction('notebook.rmt', [text], {});
            const data = result.data['application/json'];
            const textarea = document.getElementById('output');
            textarea.innerHTML = data.result;
        })();
    }, 600);  // 何も打たななかったら600ms秒後に送信
});
</script>
'''


def transform_nop(text):
    return text


def display_rmt(input_caption='入力',
                output_caption='出力',
                delay=600,
                transform_fn=transform_nop):
    data = dict(
        id=1,
        input_caption=input_caption,
        output_caption=output_caption,
        delay=delay,
    )
    _RMT_HTML = RMT_HTML.format(**data)
    _RMT_JS = RMT_JS.replace('600', str(delay))
    display(HTML(RMT_CSS+_RMT_HTML+_RMT_JS))
    _CACHE = {'': ''}

    def rmt(text):
        try:
            if text in _CACHE:
                text2 = _CACHE[text]
            else:
                text2 = transform_fn(text)
                _CACHE[text] = text2
            return JSON({'result': text2})
        except:
            traceback.print_exc()

    if google_colab is not None:
        google_colab.register_callback('notebook.rmt', rmt)


def transform_nop(text, cache):
    return text


def model_fn(text, cache):
    if text in cache:
        return cache[text]
    generated_text = model_generate(text, beam=1)
    cache[text] = generated_text
    return generated_text


def display_rmt(input_caption='入力',
                output_caption='出力',
                delay=600, html=True,
                transform_fn=model_fn):
    data = dict(
        id=1,
        input_caption=input_caption,
        output_caption=output_caption,
        delay=delay,
    )
    _RMT_HTML = RMT_HTML.format(**data)
    display(HTML(RMT_CSS+_RMT_HTML+RMT_JS))
    _CACHE = {'': ''}

    def rmt(text):
        try:
            text2 = transform_fn(text, _CACHE)
            if len(_CACHE) < 3:  # 接続エラーをキャッシュから消す
                for k, v in list(_CACHE.items()):
                    if 'HTTPConnectionPool' in v:
                        del _CACHE[k]
            return JSON({'result': text2})
        except:
            traceback.print_exc()

    if google_colab is not None:
        google_colab.register_callback('notebook.rmt', rmt)
